((nil . ((fill-column . 70)
         (sentence-end-double-space . t)
         (bug-reference-bug-regexp . "#\\(?2:[0-9]+\\)")
         (bug-reference-url-format . "https://gitlab.com/jabranham/system-packages/issues/%s"))))
